import requests

from .Utils import FetchError
from cachetools.func import ttl_cache

class Endpoint:
    def __init__(self):
        endpoints = "https://raw.githubusercontent.com/EarthMC-Toolkit/Toolkit-Website/main/endpoints.json"
        self.urls = self.reqJSON(endpoints)

    @ttl_cache(2, 300)
    def fetch(self, type, mapName):
        return self.reqJSON(self.urls[type][mapName])

    @staticmethod
    def reqJSON(url: str):
        req = requests.get(url)
        try: return req.json()
        except: return FetchError("Error fetching endpoint: " + url + "\nResponse content is not valid JSON!")

class OAPI:
    def __init__(self, map = "aurora"):
        self.domain = f"https://api.earthmc.net/v1/{map}"
        self.urls = {
            "towns": f"{self.domain}/towns",
            "nations": f"{self.domain}/nations",
            "residents": f"{self.domain}/residents",
        }

    @ttl_cache(2, 5) # Cache it for 5 seconds to avoid spamming API.
    def fetch_single(self, type: str, item = ''):
        return Endpoint.reqJSON(self.urls[type] + "/" + item)

    def fetch_all(self, type: str):
        return Endpoint.reqJSON(self.urls[type])


from .Utils import utils, FetchError

from .Classes import Nations
from .Classes import Players

from .Classes.OAPI import OAPI_Town, OAPI_Nation, OAPI_Player
from .DataHandler import OAPI

class Map(Nations):
    def __init__(self, mapName=''):
        self.name = mapName.lower()

        print('Initialising map -> ' + self.name)
        super().__init__(self.name)

        self.Players = Players(self.name, self.Towns)

        self.totalChunks = Map.townAreas(self.Towns.all())
        self.totalPlayers = Map.addAmounts(self.Players.residents.all(), self.Players.townless.all())

    @staticmethod
    def townAreas(towns): return int(sum(t['area'] for t in towns))

    @staticmethod
    def addAmounts(a1, a2): return int((len(a1) + len(a2)))

class Maps:
    @staticmethod
    def Aurora(): return Map('aurora')
    @staticmethod
    def Nova(): return Map('nova')

class _OfficialAPI:
    def __init__(self, map = "aurora"):
        self.api = OAPI(map)

    def town(self, name: str):
         return OAPI_Town(self.api.fetch_single('towns', name))

    def nation(self, name: str):
        return OAPI_Nation(self.api.fetch_single('nations', name))

    def player(self, name: str):
        return OAPI_Player(self.api.fetch_single('residents', name))

    # class Players:
    #     def __init__(self, api: OAPI):
    #         self.api = api
    #
    #     def all(self):
    #         playerArr = []
    #         playerList = self.api.fetch_all('residents')
    #
    #         for player in playerList['allResidents']:
    #             playerArr.append(self.api.fetch_single('residents', player))
    #
    #         return playerArr

OfficialAPI = _OfficialAPI()


from ..Utils import FetchError, utils, AutoRepr
from cachetools.func import ttl_cache

from ..DataHandler import Endpoint
endpoint = Endpoint()

class Town(AutoRepr):
    def __init__(self,
        name="", nation="No Nation",
        mayor="", area=0, x=0, z=0,
        residents=None, flags=None, colourCodes=None
    ):
        if residents is None:
            residents = []

        if flags is None:
            flags = {}

        if colourCodes is None:
            colourCodes = {}

        self.name = name
        self.nation = nation
        self.mayor = mayor
        self.area = area
        self.x = x
        self.z = z
        self.residents = residents
        self.flags = flags
        self.colourCodes = colourCodes

class towns:
    def __init__(self, mapName):
        self.mapName = mapName
        print("Created new 'towns' instance.")

    @ttl_cache(16, 120)
    def all(self):
        townsArray = []

        try: mapData = endpoint.fetch('map', self.mapName)
        except FetchError as e:
            print(e)
            return []

        markerset = mapData["sets"]['townyPlugin.markerset']
        areas = markerset['areas']
        townAreaNames = list(areas.keys())

        for i in range(len(townAreaNames)):
            town = areas[townAreaNames[i]]
            rawinfo = town["desc"].split("<br />")

            info = []

            for x in rawinfo:
                info.append(utils.striptags(x))

            if "Shop" in info[0]: continue

            mayor = info[1][7:]
            if mayor == "": continue

            split = info[0].split(" (")
            nationName = (split[1] if len(split) < 3 else split[2])[0:-1]
            residents = info[2][9:].split(", ")

            x = round((max(town["x"]) + min(town["x"])) / 2)
            z = round((max(town["z"]) + min(town["z"])) / 2)

            flags = {
                'pvp': utils.strAsBool(info[4][5:]),
                'mobs': utils.strAsBool(info[5][6:]),
                'public': utils.strAsBool(info[6][8:]),
                'explosions': utils.strAsBool(info[7][11:]),
                'fire': utils.strAsBool(info[8][6:]),
                'capital': utils.strAsBool(info[9][9:])
            }

            colourCodes = {
                'fill': town['fillcolor'],
                'outline': town['color']
            }

            ct = Town(
                name=utils.removeStyleCharacters(town['label']),
                mayor=mayor,
                area=utils.townArea(town),
                x=x, z=z,
                residents=residents,
                flags=flags,
                colourCodes=colourCodes
            )

            if nationName != '':
                ct.nation = nationName.strip()

            townsArray.append(ct)

        cachedTowns = []
        temp = {}

        for t in townsArray:
            if temp.get(t.name, None) is None:
                temp[t.name] = t
                cachedTowns.append(vars(t))
            else:
                temp[t.name].area += t.area

        return cachedTowns

    def get(self, townName, towns=None):
        if towns is None: towns = self.all()
        foundTown = utils.find(lambda town: town['name'].lower() == townName.lower(), towns)

        if foundTown is None: return "Could not find town '" + townName + "'"
        return foundTown


from EarthMC.Utils import AutoRepr

class OAPI_Town(AutoRepr):
    def __init__(self, data):
        timestamps = data['timestamps']
        status = data['status']
        stats = data['stats']

        self.timestamps = {
            'founded': timestamps['registered'],
            'joinedNation': timestamps['joinedNationAt']
        }

        strings = data['strings']
        self.name = strings['board']
        self.board = strings['board']
        self.colour = strings['mapColorHexCode']
        self.mayor = strings['mayor']
        self.founder = strings['founder']
        self.nation = data['affiliation']['nation']

        self.area = stats['numTownBlocks']
        self.x = data['home']['x']
        self.z = data['home']['z']

        self.residents = data['residents']
        self.spawn = data['spawn']

        perms = data['perms']
        rnaoPerms = perms['rnaoPerms']

        self.perms = {
            'switch': rnaoPerms['switchPerms'],
            'itemUse': rnaoPerms['itemUsePerms'],
            'build': rnaoPerms['buildPerms'],
            'destroy': rnaoPerms['destroyPerms']
        }

        self.flags = perms['flagPerms']
        self.ranks = data['ranks']


from ..Utils import utils, AutoRepr

from .Towns import towns
from cachetools.func import ttl_cache

class Nation(AutoRepr):
    def __init__(self, name="", king="", capital="", residents=None, towns=None, area=0):
        if towns is None:
            towns = []

        if residents is None:
            residents = []

        self.name = name
        self.king = king
        self.capital = capital
        self.residents = residents
        self.towns = towns
        self.area = area

class nations:
    def __init__(self, mapName):
        self.Towns = towns(mapName)
        self.Nations = self

        print("Created new 'nations' instance")

    @ttl_cache(4, 120)
    def get(self, nationName, nations=None):
        if nations is None:
            nations = self.all()

        foundNation = utils.find(lambda n: n['name'].lower() == nationName.lower(), nations)
        if foundNation is None:
            return "Could not find nation '" + nationName + "'"

        return foundNation

    @ttl_cache(16, 120)
    def all(self):
        raw = {}
        output = []

        for town in self.Towns.all():
            nationName = town["nation"]
            if nationName == 'No Nation':
                continue

            # Doesn't already exist, create new nation.
            if raw.get(nationName, None) is None:
                raw[nationName] = Nation(
                    name=nationName,
                    residents=town['residents'],
                    towns=[],
                    area=0
                )

                output.append(vars(raw[nationName]))

            # Add up existing values
            townName = town['name']
            raw[nationName].area += town['area']

            raw[nationName].residents.extend(town['residents'])
            raw[nationName].residents = utils.listFromDictKey(raw[nationName].residents)

            if raw[nationName].name == town['nation']:
                raw[nationName].towns.append(townName)

            if town['flags']['capital'] is True:
                raw[nationName].king = town['mayor']
                raw[nationName].capital = {
                    'name': townName,
                    'x': town['x'],
                    'z': town['z']
                }

        return output


from EarthMC.Utils import AutoRepr

class OAPI_Player(AutoRepr):
    def __init__(self, data):
        self.name = data['name']
        self.nation = data['nation']
        self.town = data['town']
        self.balance = data['balance']
        self.registered = data['registered']


import re

from io import StringIO
from html.parser import HTMLParser

class AutoRepr(object):
    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join('{}={!r}'.format(k, v) for k, v in self.__dict__.items())
        )

class MLStripper(HTMLParser):
    def __init__(self):
        super().__init__()
        self.reset()
        self.strict = False
        self.convert_charrefs = True
        self.text = StringIO()
    def handle_data(self, d): self.text.write(d)
    def get_data(self): return self.text.getvalue()

class FetchError(Exception):
    """Raised when there was an error fetching data from Dynmap."""
    pass

class utils:
    @staticmethod
    def striptags(html):
        s = MLStripper()
        s.feed(html)
        return s.get_data()

    @staticmethod
    def find(pred, iterable):
        found = None
        for element in iterable:
            if pred(element):
                found = element
                break

        return found

    # @staticmethod
    # def intersection(arr1, arr2): return list(filter(lambda x: x in arr1, arr2))
    # def difference(arr1, arr2): return filter(lambda x: x not in arr1, arr2)

    @staticmethod
    def removeStyleCharacters(string): return re.sub('/(&amp;.|&[0-9kmnola-z])/g', "", string)

    @staticmethod
    def strAsBool(string): return True if string == 'true' else False

    @staticmethod
    def parseObject(obj): return vars(obj) if type(obj) is not dict else obj
    @staticmethod
    def dictToList(dict):
        vals = dict.values()
        return [utils.parseObject(val) for val in vals]

    @staticmethod
    def listFromDictKey(key): return list(dict.fromkeys(key))

    @staticmethod
    def townArea(town): return utils.calcArea(town['x'], town['z'], len(town['x']))

    @staticmethod
    def calcArea(x, z, points, divisor=256):
        area = 0
        j = points-1

        for i in range(points):
            area += (x[j] + x[i]) * (z[j] - z[i])
            j = i

        return abs(area/2) / divisor